/*
 filter.c written by ksr
 Continuous preprocessing handlers of http stream.
 http is used for various purpose.  
 So not only rendering html but other handlers are are required.

 I don't know where zlib transfer encoding should be handled.

 raw data comes from network 
   -> zlib(gzip) 
   -> convert kanji to internal encoding to be displayed.

 User can add user defined preprocess stream.  For example,
 
 raw data from network 
   -> zlib(gzip) 
   -> convert kanji to internal encoding to be displayed.
   -> user defined preprocess ( language translater etc )
   -> rendering html
 */
#include <gtk/gtk.h>
#include <sys/types.h>
#include "filter.h"
#if ZLIB
#include "zlib.h"
#endif
#include "dillo.h"

void a_Filter_init(DilloFilter *p)
{
  p->malloc = (void *(*)(size_t))g_malloc;
  p->realloc = (void *(*)(void *, size_t))g_realloc;
  p->free = (void (*)(void *))g_free;
  p->inbufsize = 0;
  p->inbuf = 0;
  p->inpos = 0;
  p->total_in = 0;
  p->outbufsize = 0;
  p->outbuf = 0;
  p->total_out = 0;
  p->outpos = 0;
  p->ext = 0;
  p->extfree = 0;
  p->flag = 0;
}

void a_Filter_set_in(DilloFilter *p, unsigned char *buf, size_t bufsize, size_t newpos)
{
    if (p->flag & 1) {
      p->free(p->inbuf);
    }
    p->inbufsize = bufsize;
    p->total_in = bufsize;
    p->inbuf = buf;
}

void a_Filter_append_input(DilloFilter *p, unsigned char *buf, size_t bufsize)
{
  if ((p->flag & 1) == 0 && p->inbuf == 0) { abort(); }
  if (p->inbuf == 0) {
    p->inbufsize = 4096;
    p->inbuf = p->malloc(p->inbufsize+1);
    p->flag |= 1;
  }
  if (p->total_in + bufsize > p->inbufsize) {
    p->inbufsize += 4096;
    p->inbuf = p->realloc(p->inbuf, p->inbufsize+1);
  }
  memmove(&p->inbuf[p->total_in], buf, bufsize);
  p->total_in += bufsize;
}

void a_Filter_append_output(DilloFilter *p, unsigned char *buf, size_t bufsize)
{
  if (p->outbuf == 0) {
    p->outbufsize = 4096;
    p->outbuf = p->malloc(p->outbufsize);
    p->flag |= 2;
  }
  if (p->total_out + bufsize >= p->outbufsize) {
    p->outbufsize += (bufsize * 2);
    p->outbuf = p->realloc(p->outbuf, p->outbufsize);
  }
  memmove(&p->outbuf[p->total_out], buf, bufsize);
  p->total_out += bufsize;
}


int a_Filter_process(DilloFilter *p, int eof)
{
  unsigned char *in;
  size_t insize;
  unsigned char *out;
  size_t outsize;
  int ret;
  in = &p->inbuf[p->inpos];
  insize = p->total_in - p->inpos;
  out = &p->outbuf[p->outpos];
  outsize = p->outbufsize - p->outpos;
  ret = p->process(p, &in, &insize, &out, &outsize, eof);
  if (ret == -1)
    return -1;
  if (ret) {
    /* I can believe in returning value only */
    p->inpos += (in - &p->inbuf[p->inpos]);
    p->outpos += ret;
    if (p->total_out < p->outpos)
      p->total_out = p->outpos;
  }
  return ret;
}

void a_Filter_free(DilloFilter *p)
{
  if (p->extfree) p->extfree(p->ext);
  if (p->flag & 1) p->free(p->inbuf);
  if (p->flag & 2) p->free(p->outbuf);
  g_free(p);
}

#if ZLIB
static int process_decompress(DilloFilter ps, unsigned char **inptr, size_t *insize, unsigned char **outptr, size_t *outsize, int eof)
{
  z_stream *zs = ps->ext;
  int status;
  zs->next_in = *inptr;
  zs->avail_in = *insize;
  zs->next_out = *outptr;
  zs->avail_out = *outsize;
  status = inflate(zs);
  if (status == Z_STREAM_END) {
    return 0;
  }
  if (status != Z_OK) {
    inflateEnd(zs);
    return -1;
  }
  *inptr = zs->next_in;
  ret = *outsize - zs->avail_out;
  *insize = zs->avail_in;
  *outptr = zs->nextout;
  *outsize = zs->avail_out;
  if (eof)
    inflateEnd(zs);
  return ret;
}
#endif

/*
                 IN             OUT
  inptr         input        -> input + converted
  insize        input length -> left
  outptr        output       -> output + outputted len
  outsize       output buffer length -> left
 */
gboolean Html_guess_encoding(DilloHtml *html, guchar *buf, gint bufsize,
			     gint Eof) ;

static int process_translate_encoding(DilloFilter *ps, unsigned char **inptr, size_t *insize, unsigned char **outptr, size_t *outsize, int eof)
{
  DilloHtml *html = ps->ext;
  int output;
  if (!Html_guess_encoding(html, ps->inbuf, ps->total_in, eof)) return 0;

  if (html->trans) {
    int previn = html->trans->converted;
    int prevout = html->trans->bufsize;
    Html_translate(html->trans, ps->inbuf, ps->total_in, eof);
    ps->outbuf = html->trans->buffer;
    ps->outbufsize = html->trans->bufsize;
    output = html->trans->bufsize - prevout;
    *outsize = 0;
    *outptr = html->trans->buffer + html->trans->bufsize;
    *insize = 0;
    *inptr = ps->inbuf + html->trans->converted;
    return output;
  } else {
    output = *insize;
    a_Filter_append_output(ps, *inptr, *insize);
    *outptr = ps->outbuf + ps->total_out;
    *outsize = *outsize - *insize;
    *inptr += *insize;
    *insize = 0;
    return output;
  }
}

DilloFilter *a_Filter_get_translate_filter(void *ext)
{
  DilloFilter *p = g_malloc(sizeof(DilloFilter));
  DilloHtml *html = ext;
  a_Filter_init(p);
  p->process = process_translate_encoding;
  p->ext = html;
  
  return p;
}

#if ZLIB
DilloFilter *a_Filter_get_decompress_filter(void *ext)
{
  DilloFilter *p = g_malloc(sizeof(DilloFilter));
  DilloHtml *html = ext;
  z_stream *z;
  a_Filter_init(p);
  p->process = process_decompress;
  p->ext = z = g_malloc(sizeof(z_stream));
  z->zalloc = 0;
  z->zfree = 0;
  z->opaque = 0;
  return p;
}
#endif


