/*
 * File: intrinsic_iconv.c
 *
 * Copyright (c) 2001 ksr <ksr@todo.org>
 *
 * BSD license without the requirement of indicating copyright notice, 
 * but you have to keep it in the source.
 */

#include "intrinsic_iconv.h"
#include <errno.h>
#include <string.h>
#include "debug.h"

static unsigned char *ustrchr(unsigned char *p, int ch)
{
  for (;;++p) {
    if (*p == ch) return p;
    if (!*p) return 0;
  }
}

static int sjis2eucjp(iconvjp_t cd, unsigned char **_in, size_t *_insize,
		       unsigned char **_out, size_t *_outsize)
{
  unsigned char *in = *_in;
  unsigned char *out = *_out;
  size_t insize = *_insize;
  size_t outsize = *_outsize;
  unsigned char *in_end = in + insize;
  unsigned char *out_end = out + outsize;

  for ( ; in < in_end; )
  {
    int ch = *in;
    if ( ch <= 0x7f )
    {
   /* JIS-X-0201 */
      *out++ = *in++;
    }
    else if ( ch >= 0xA1 && ch <= 0xDF )
    {
      /* hankaku katakana */
#define KATAKANA_H2Z 1
#if KATAKANA_H2Z
      int i;
      static unsigned char z[] = "\xa1\xa1\xa1\xa3\xa1\xd6\xa1\xd7\xa1\xa2\xa1\xa6\xa5\xf2\xa5\xa1\xa5\xa3\xa5\xa5\xa5\xa7\xa5\xa9\xa5\xe3\xa5\xe5\xa5\xe7\xa5\xc3\xa1\xbc\xa5\xa2\xa5\xa4\xa5\xa6\xa5\xa8\xa5\xaa\xa5\xab\xa5\xad\xa5\xaf\xa5\xb1\xa5\xb3\xa5\xb5\xa5\xb7\xa5\xb9\xa5\xbb\xa5\xbd\xa5\xbf\xa5\xc1\xa5\xc4\xa5\xc6\xa5\xc8\xa5\xca\xa5\xcb\xa5\xcc\xa5\xcd\xa5\xce\xa5\xcf\xa5\xd2\xa5\xd5\xa5\xd8\xa5\xdb\xa5\xde\xa5\xdf\xa5\xe0\xa5\xe1\xa5\xe2\xa5\xe4\xa5\xe6\xa5\xe8\xa5\xe9\xa5\xea\xa5\xeb\xa5\xec\xa5\xed\xa5\xef\xa5\xf3\xa1\xab\xa1\xac";
      static unsigned char dakuh[] = "\xb3\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xca\xcb\xcc\xcd\xce";
      static unsigned char dakuz[] = "\xa5\xf4\xa5\xac\xa5\xae\xa5\xb0\xa5\xb2\xa5\xb4\xa5\xb6\xa5\xb8\xa5\xba\xa5\xbc\xa5\xbe\xa5\xc0\xa5\xc2\xa5\xc5\xa5\xc7\xa5\xc9\xa5\xd0\xa5\xd3\xa5\xd6\xa5\xd9\xa5\xdc";
      static unsigned char handakuh[] = "\xca\xcb\xcc\xcd\xce";
      static unsigned char handakuz[] = "\xa5\xd1\xa5\xd4\xa5\xd7\xa5\xda\xa5\xdd";

      if (in_end - in < 2 && !cd->compat) 
	goto end;
      if (*(in+1) == 0xde) {           /* DAKUON */
	unsigned char *found = ustrchr(dakuh, *in);
	if (found) {
	  i = (found - (unsigned char *)dakuh) * 2;
	  *out++ = dakuz[i];
	  *out++ = dakuz[i+1];
	}
	in += 2;
      } else if (*(in+1) == 0xdf) {    /* HANDAKUON */
	unsigned char *found = ustrchr(handakuh, *in);
	if (found) {
	  i = (found - (unsigned char *)handakuh) * 2;
	  *out++ = handakuz[i];
	  *out++ = handakuz[i+1];
	}
	in += 2;
      } else {
	i = (*in - 0xa0) * 2;
	*out++ = z[i];
	*out++ = z[i+1];
	in++;
      }
#else
      *out++ = '\x8E';
      *out++ = *in++;
#endif
    }
    else
    {
      /* JIS-X-0208 */
      int lo;
      int hi;
      if (in_end - in < 2 && !cd->compat) 
	goto end;

      lo = ch - ( ch <= 0x9F ? 0x71 : 0xB1 );
      hi = *(in+1);
      
      lo = (lo << 1) + 1;
      
      if (hi > 0x7F)
        hi--;
      if ( hi < 0x9E )
        hi -= 0x1F;
      else
      {
        hi -= 0x7D;
        lo++;
      }
      *out++ = lo + 0x80;
      *out++ = hi + 0x80;
      in += 2;
    }
  }
end:
  *_insize = in_end - in;
  *_outsize = out_end - out;
  *_in = in;
  *_out = out;
  return 1;
}

#define us_ascii 0                   /* US ASCII */
#define jisx0201_roman 1             /* JIS X 0201 Roman */
#define jisx0201_kana 2              /* JIS X 0201 Katakana */
#define jisx0208_1983 3              /* JIS X 0208-1983 */
#define jisx0208_1990 4              /* JIS X 0208-1990  */
#define jisx0212_1990 5              /* JIS X 0212-1990 */
#define jisx0208_1976 6              /* JIS X 0212-1976 */

static
int iso2022jp2eucjp(iconvjp_t cd, unsigned char **_in, size_t *_insize,
		     unsigned char **_out, size_t *_outsize)
{
  unsigned char *in = *_in;
  unsigned char *out = *_out;
  size_t insize = *_insize;
  size_t outsize = *_outsize;
  unsigned char *in_end = in + insize;
  unsigned char *out_end = out + outsize;
  int state = cd->state;

  for (; in < in_end; )
  {
    if ( in[0] == 0x1B ) {
      if (in[1] == 0x28 &&
	  in[2] == 0x42 )
	{
	  /* US ASCII */
	  state = us_ascii;
	  in += 3;
	  continue;
	}
      if (in[1] == 0x28)
	{
      /* JIS X 0201 Roman */
	  if (in[2] == 0x4A) {
	    state = jisx0201_roman;
	    in += 3;
	    continue;
          }
	  if (in[2] == 0x49 ) {
      /* JIS X 0201 Katakana */
	    state = jisx0201_kana;
	    in += 3;
	    continue;
  	  }
	}
      if (in[1] == 0x24) {
	  if (in[2] == 0x42 ) {
      /* JIS X 0208-1983 */
 	    state = jisx0208_1983;
	    in += 3;
	    continue;
          }
    // JIS X 0208-1976
          if ( in[2] == 0x40 ) {
	    state = jisx0208_1976;
	    in += 3;
	    continue;
	  }
	}
      if (in[1] == 0x26 &&
	  in[2] == 0x40 &&
	  in[3] == 0x1B &&
	  in[4] == 0x24 &&
	  in[5] == 0x42 )
	{
      /* JIS X 0208-1990 */
	  state = jisx0208_1990;
	  in += 6;
	  continue;
	}
    // JIS X 0212-1990
      if ( in[1] == 0x24 &&
	   in[2] == 0x28 &&
	   in[3] == 0x44 )
	{
	  state = jisx0212_1990;
	  in += 4;
	  continue;
	}
      if (in_end - in < 3 && !cd->compat) {
	goto end;
      }
    }
    switch ( state )
    {
    case us_ascii:
    case jisx0201_roman:
      *out ++ = *in++;
      break;
    case jisx0201_kana:
      *out++ = 0x8e;
      *out++ = *in++ + 0x80;
      break;
    case jisx0208_1976:
    case jisx0208_1983:
    case jisx0208_1990:
      if (in_end - in < 2) goto end;
      *out++ = *in++ + 0x80;
      *out++ = *in++ + 0x80;
      break;
    case jisx0212_1990:
      if (in_end - in < 2) goto end;
      *out++ = 0x8f;
      *out++ = *in++ + 0x80;
      *out++ = *in++ + 0x80;
      break;
    default:
      ;
    }
  }
 end:
  cd->state = state;
  *_insize = in_end - in;
  *_outsize = out_end - out;
  *_out = out;
  *_in = in;
  return 1;
}


iconvjp_t iconvjp_open (char *dstset, char *srcset)
{
  iconvjp_t p;
  p = (iconvjp_t)g_malloc(sizeof(struct iconvjp));
  p->state = 0;
  p->compat = 0;
  if (!g_strcasecmp(srcset, "ISO-2022-JP")) {
    p->convert = iso2022jp2eucjp;
  } else if (!g_strcasecmp(srcset, "EUCJP")) {
    goto error;
  } else if (!g_strcasecmp(srcset, "SJIS")) {
    p->convert = sjis2eucjp;
  } else {
    goto error;
  }
  if (!g_strcasecmp(dstset, "EUCJP")) {
    ;
  } else {
    goto error;
  }

  return p;
 error:
  errno = EINVAL;
  return (void*)-1;
}

void *iconvjp (iconvjp_t cd, char **inptr, size_t *insize, char **outptr, size_t *outsize)
{
  if (inptr == 0) {
    /* set to initial shift state */
    cd->state = 0;
    return 0;
  }
  if (*cd->convert == 0) {
    int osize = *insize;
    memmove(*outptr, *inptr, osize);
    *outptr = *outptr + osize;
    *outsize = osize;
    return (void *)osize;
  }
  if (!cd->convert(cd, (unsigned char **)inptr, insize, (unsigned char **)outptr, outsize))
    goto error;
  return 0;
 error:
  errno = EINVAL;
  return (void *)-1;
}

void iconvjp_close(iconvjp_t cd)
{
  g_free(cd);
}

void iconvjp_compat(iconvjp_t cd, int flag)
{
  cd->compat = flag;
}
